from dataclasses import dataclass
from typing import Dict, List, ClassVar
from pyspark.sql import DataFrame, functions as F
from .engine_rules import register_inspection_rule, InspectionRule
from .engine_reporting import InspectionResult

# Rule-specific Parameters


@dataclass(frozen=True)
class PrivacyParams:
    source_data: Dict[str, DataFrame]
    synth_data: Dict[str, DataFrame]
    sensitive_columns: Dict[str, List[str]]


# Structural Rules
@register_inspection_rule("check_schema_structure")
class ColumnMatchesCheck(InspectionRule[PrivacyParams]):

    params: ClassVar[type[PrivacyParams]] = PrivacyParams

    def inspect(self) -> InspectionResult:
        """
        Compare synthetic dataset back to source dataset, checking for matches
        to production values
        """
        source_data = self.params.source_data
        synth_data = self.params.synth_data
        sensitive_columns = self.params.sensitive_columns

        theme = "Privacy"
        name = "check_column_level_matches"

        matched_columns: List[str] = []

        for key, source_df in source_data.items():
            print(f"insepcting table {key}")
            synth_df = synth_data[key]

            cols_to_inspect = sensitive_columns.get(key, [])
            print(cols_to_inspect)
            if not cols_to_inspect:
                print(f"no cols to inspect in {key}")
                continue

            for col_name in cols_to_inspect:

                matched_values = (
                    source_df.select(F.col(col_name))
                    .distinct()
                    .intersect(synth_df.select(F.col(col_name)).distinct())
                )

                if matched_values.count() > 0:
                    matched_columns.append(f"{key}.{col_name}")
                    status = "Failed"
                    message = f"Sensitive value match detected in {matched_columns[0]}"
                    return InspectionResult(
                        theme=theme, name=name, status=status, message=message
                    )

        status = "Passed"
        message = "OK"
        return InspectionResult(theme=theme, name=name, status=status, message=message)



from dataclasses import dataclass, field, asdict
from typing import Dict, Any, Union
from pyspark.sql import DataFrame

from .profiler_table import TableProfile, profile_table


@dataclass(slots=True)
class SchemaProfile:
    tables: Dict[str, TableProfile] = field(default_factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


class SchemaProfiler:
    def __init__(self, default_single_name: str = "table") -> None:
        self._default_single_name = default_single_name

    def profile_schema(
        self, tables: Union[DataFrame, Dict[str, DataFrame]]
    ) -> SchemaProfile:
        table_map = self._as_table_map(tables, self._default_single_name)

        table_profiles: Dict[str, TableProfile] = {}
        for table_name, df in table_map.items():
            table_profiles[table_name] = profile_table(df, table_name)

        return SchemaProfile(tables=table_profiles)

    @staticmethod
    def _as_table_map(
        tables: Union[DataFrame, Dict[str, DataFrame]], default_single_name: str
    ) -> Dict[str, DataFrame]:
        if isinstance(tables, DataFrame):
            return {default_single_name: tables}
        if isinstance(tables, dict):
            for k, v in tables.items():
                if not isinstance(k, str) or not isinstance(v, DataFrame):
                    raise TypeError("Expected Dict[str, DataFrame]")
            return tables
        raise TypeError("Expected a DataFrame or Dict[str, DataFrame]")



from dataclasses import dataclass
from datetime import date
from typing import Dict, Any, Optional
from abc import ABC, abstractmethod

from pyspark.sql import DataFrame, functions as F, types as T


# Parent Base Class
@dataclass(slots=True)
class BaseColumnProfile(ABC):
    name: str
    normalised_type: str
    nullable: bool

    @abstractmethod
    def default_rule(self) -> str: ...

    @abstractmethod
    def type_specific_params(self) -> Dict[str, Any]: ...


# Date Type Profiling
@dataclass(slots=True)
class DateColumnProfile(BaseColumnProfile):
    min_date: Optional[date] = None
    max_date: Optional[date] = None

    def default_rule(self) -> str:
        return "random_date"

    def type_specific_params(self) -> Dict[str, Any]:
        return {
            "min_date": self.min_date.isoformat() if self.min_date else None,
            "max_date": self.max_date.isoformat() if self.max_date else None,
        }


def profile_date_column(df: DataFrame, col_name: str) -> DateColumnProfile:
    field = df.schema[col_name]
    nullable = field.nullable

    col_profile = (
        df.select(F.col(col_name).alias("val")).agg(
            F.min("val").alias("min_date"),
            F.max("val").alias("max_date"),
        )
    ).first()

    col_stats = col_profile.asDict() if col_profile else {}

    return DateColumnProfile(
        name=col_name,
        normalised_type="date",
        nullable=nullable,
        min_date=col_stats.get("min_date"),
        max_date=col_stats.get("max_date"),
    )


# Numeric Type Profiling
@dataclass(slots=True)
class NumericColumnProfile(BaseColumnProfile):
    min_value: Optional[float] = None
    max_value: Optional[float] = None
    mean_value: Optional[float] = None

    def default_rule(self) -> str:
        return "random_int"

    def type_specific_params(self) -> Dict[str, Any]:
        return {
            "min_value": self.min_value,
            "max_value": self.max_value,
            "mean_value": self.mean_value,
        }


def profile_numerical_column(df: DataFrame, col_name: str) -> NumericColumnProfile:
    field = df.schema[col_name]
    nullable = field.nullable

    col_profile = (
        df.select(F.col(col_name).alias("val")).agg(
            F.min("val").alias("min_value"),
            F.max("val").alias("max_value"),
            F.avg("val").alias("mean_value"),
        )
    ).first()

    col_stats = col_profile.asDict() if col_profile else {}

    return NumericColumnProfile(
        name=col_name,
        normalised_type="numeric",
        nullable=nullable,
        min_value=col_stats.get("min_value"),
        max_value=col_stats.get("max_value"),
        mean_value=col_stats.get("mean_value"),
    )


# String Type Profiling


@dataclass(slots=True)
class StringColumnProfile(BaseColumnProfile):
    min_length: Optional[int] = None
    max_length: Optional[int] = None
    mean_length: Optional[float] = None

    def default_rule(self) -> str:
        return "random_string"

    def type_specific_params(self) -> Dict[str, Any]:
        return {
            "min_length": self.min_length,
            "max_length": self.max_length,
            "mean_length": self.mean_length,
        }


def profile_string_column(df: DataFrame, col_name: str) -> StringColumnProfile:
    field = df.schema[col_name]
    nullable = field.nullable

    length_col = F.length(F.col(col_name))

    col_profile = (
        df.select(length_col.alias("len")).agg(
            F.min("len").alias("min_length"),
            F.max("len").alias("max_length"),
            F.avg("len").alias("mean_length"),
        )
    ).first()

    col_stats = col_profile.asDict() if col_profile else {}

    return StringColumnProfile(
        name=col_name,
        normalised_type="string",
        nullable=nullable,
        min_length=col_stats.get("min_length"),
        max_length=col_stats.get("max_length"),
        mean_length=col_stats.get("mean_length"),
    )



from abc import ABC, abstractmethod
from typing import Any, Dict, Type, Mapping, TypeVar, Generic, ClassVar
from .engine_reporting import InspectionResult

ParamsType = TypeVar("ParamsType")


class InspectionRule(Generic[ParamsType], ABC):
    """ "
    This is the base rule for all inspection rules.

    All subclasses (specific inspection rules) to set:
        PARAMS_MODEL: The rule-specific params dataclass type
    """

    PARAMS_TYPE: ClassVar[type]

    def __init__(self, params: ParamsType):
        self.params: ParamsType = params

    @abstractmethod
    def inspect(self) -> InspectionResult: ...


INSPECTION_RULES_REGISTRY: Dict[str, Type[InspectionRule[Any]]] = {}


def register_inspection_rule(rule_name: str):
    """
    Decorator to register an inspection rule class
    """

    def wrapper(rule_class):
        INSPECTION_RULES_REGISTRY[rule_name] = rule_class
        return rule_class

    return wrapper


def get_inspection_rule(
    rule_name: str, params_dict: Mapping[str, Any]
) -> InspectionRule:
    """
    Factory to instantiate a rule by name
    """
    if rule_name not in INSPECTION_RULES_REGISTRY:
        raise ValueError(f"Rule {rule_name} is not registered")

    rule_class = INSPECTION_RULES_REGISTRY[rule_name]
    rule_params_type: Type[Any] = rule_class.PARAMS_TYPE
    rule_params = rule_params_type(**params_dict)
    return rule_class(rule_params)



from typing import Dict, Any
import hashlib
from respark.planning import (
    SchemaGenerationPlan,
    TableGenerationPlan,
    get_generation_rule,
)
from pyspark.sql import SparkSession, DataFrame, Column, functions as F, types as T


def _create_stable_seed(base_seed: int, *tokens: Any) -> int:
    payload = "|".join([str(base_seed), *map(str, tokens)]).encode("utf-8")
    digest = hashlib.sha256(payload).digest()
    val64 = int.from_bytes(digest[:8], byteorder="big", signed=False)
    mixed = val64 ^ (base_seed & 0x7FFFFFFFFFFFFFFF)

    return mixed & 0x7FFFFFFFFFFFFFFF


def _str_to_spark_type(type_as_str: str) -> T.DataType:
    if type_as_str == "numeric":
        return T.IntegerType()
    if type_as_str == "string":
        return T.StringType()
    if type_as_str == "date":
        return T.DateType()
    else:
        raise TypeError("Unsupported Type")


class SynthSchemaGenerator:
    def __init__(self, spark: SparkSession, seed: int = 18151210):
        self.spark = spark
        self.seed = int(seed)

    def generate_synthetic_schema(
        self, spark: SparkSession, schema_gen_plan: SchemaGenerationPlan
    ) -> Dict[str, DataFrame]:

        synth_schema: Dict[str, DataFrame] = {}

        for table_plan in schema_gen_plan.tables:
            table_generator = SynthTableGenerator(
                spark_session=self.spark,
                table_gen_plan=table_plan,
                seed=self.seed,
            )
            synth_schema[table_generator.table_gen_plan.name] = (
                table_generator.generate_synthetic_table()
            )

        return synth_schema


class SynthTableGenerator:
    def __init__(
        self,
        spark_session: SparkSession,
        table_gen_plan: TableGenerationPlan,
        seed: int = 18151210,
    ):
        self.spark = spark_session
        self.table_gen_plan = table_gen_plan
        self.table_name = table_gen_plan.name
        self.row_count = table_gen_plan.row_count
        self.seed = seed

    def generate_synthetic_table(self):
        synth_df = self.spark.range(0, self.row_count, 1)
        synth_df = synth_df.withColumnRenamed("id", "__row_idx")

        for column_plan in self.table_gen_plan.columns:
            col_seed = _create_stable_seed(
                self.seed, self.table_name, column_plan.name, column_plan.rule
            )

            exec_params = {
                **column_plan.params,
                "__seed": col_seed,
                "__table": self.table_name,
                "__column": column_plan.name,
                "__dtype": column_plan.data_type,
                "__row_idx": F.col("__row_idx"),
            }

            rule = get_generation_rule(column_plan.rule, **exec_params)
            col_expr: Column = rule.generate_column()

            target_dtype = _str_to_spark_type(column_plan.data_type)
            col_expr = col_expr.cast(target_dtype)

            synth_df = synth_df.withColumn(column_plan.name, col_expr)

        ordered_cols = [cgp.name for cgp in self.table_gen_plan.columns]
        return synth_df.select("__row_idx", *ordered_cols).drop("__row_idx")



from .profiler_column import (
    BaseColumnProfile,
    StringColumnProfile,
    NumericColumnProfile,
    DateColumnProfile,
    profile_string_column,
    profile_numerical_column,
    profile_date_column,
)
from .profiler_table import TableProfile, profile_table
from .profiler_schema import SchemaProfile, SchemaProfiler



from typing import Any, List, Union
from pyspark.sql import Column, functions as F


class RNG:
    def __init__(self, row_idx: Column, base_seed: int):
        self.row_idx = row_idx
        self.seed = int(base_seed)
        self._U53 = float(1 << 53)
        self._U53_INT = 1 << 53

    def _u64(self, *salt: Any) -> Column:

        parts: List[Column] = [F.lit(self.seed)]
        for s in salt:
            parts.append(s if isinstance(s, Column) else F.lit(str(s)))
        parts.append(self.row_idx)
        return F.xxhash64(*parts)

    def u01(self, *salt: Any) -> Column:
        return (F.pmod(self._u64(*salt), F.lit(self._U53_INT)) / F.lit(self._U53)).cast(
            "double"
        )

    def randint(self, low: int, high: int, *salt: Any) -> Column:
        span = int(high) - int(low) + 1
        return (F.pmod(self._u64(*salt), F.lit(span)) + F.lit(int(low))).cast("int")

    def choice(self, options: Union[List[Any], Column], *salt: Any) -> Column:
        if isinstance(options, Column):
            arr = options
            arr_len_col = F.size(arr)
            idx1 = (F.pmod(self._u64(*salt), arr_len_col) + F.lit(1)).cast("int")
            return F.element_at(arr, idx1)
        else:
            arr = F.array([F.lit(v) for v in options])
            idx1 = (F.pmod(self._u64(*salt), F.lit(len(options))) + F.lit(1)).cast(
                "int"
            )
            return F.element_at(arr, idx1)



import string
from abc import ABC, abstractmethod
from datetime import datetime
from pyspark.sql import Column, functions as F, types as T
from .random_helpers import RNG


class GenerationRule(ABC):
    def __init__(self, **params):
        self.params = params

    @property
    def seed(self) -> int:
        return int(self.params["__seed"])

    @property
    def row_idx(self) -> Column:
        return self.params["__row_idx"]

    def rng(self) -> RNG:
        return RNG(self.row_idx, self.seed)

    @abstractmethod
    def generate_column(self) -> Column:
        pass


GENERATION_RULES_REGISTRY = {}


def register_generation_rule(rule_name: str):
    """
    Decorator to register a generation rule class
    """

    def wrapper(cls):
        GENERATION_RULES_REGISTRY[rule_name] = cls
        return cls

    return wrapper


def get_generation_rule(rule_name: str, **params) -> GenerationRule:
    """
    Factory to instantiate a rule by name
    """
    if rule_name not in GENERATION_RULES_REGISTRY:
        raise ValueError(f"Rule {rule_name} is not registered")
    return GENERATION_RULES_REGISTRY[rule_name](**params)


# Date Rules
@register_generation_rule("random_date")
class RandomDateRule(GenerationRule):
    def generate_column(self) -> Column:
        min_date_str = self.params.get("min_date", "2000-01-01")
        max_date_str = self.params.get("max_date", "2025-12-31")

        min_date = datetime.strptime(min_date_str, "%Y-%m-%d")
        max_date = datetime.strptime(max_date_str, "%Y-%m-%d")
        days_range = (max_date - min_date).days

        rng = self.rng()
        offset = rng.randint(0, days_range)
        return F.date_add(F.to_date(F.lit(min_date_str)), offset).cast(T.DateType())


# Numeric Rules
@register_generation_rule("random_int")
class RandomIntRule(GenerationRule):
    def generate_column(self) -> Column:
        min_value = self.params.get("min_value", 0)
        max_value = self.params.get("max_value", 2147483647)

        rng = self.rng()
        col = rng.randint(min_value, max_value)
        return col.cast(T.IntegerType())


# String Rules
@register_generation_rule("random_string")
class RandomStringRule(GenerationRule):
    def generate_column(self) -> Column:
        min_length = self.params.get("min_length", 0)
        max_length = self.params.get("max_length", 50)
        charset = self.params.get("charset", string.ascii_letters)

        rng = self.rng()

        length = rng.randint(min_length, max_length, "len")
        charset_arr = F.array([F.lit(c) for c in charset])

        pos_seq = F.sequence(F.lit(0), F.lit(max_length - 1))
        chars = F.transform(pos_seq, lambda p: rng.choice(charset_arr, "pos", p))

        return F.concat_ws("", F.slice(chars, 1, length)).cast(T.StringType())



from dataclasses import dataclass
from typing import Dict, ClassVar
from pyspark.sql import DataFrame
from .engine_rules import register_inspection_rule, InspectionRule
from .engine_reporting import InspectionResult

# Rule-specific Parameters


@dataclass(frozen=True)
class SchemaStructureParams:
    source_data: Dict[str, DataFrame]
    synth_data: Dict[str, DataFrame]


# Structural Rules
@register_inspection_rule("check_schema_structure")
class SchemaStructureCheck(InspectionRule[SchemaStructureParams]):

    params: ClassVar[type[SchemaStructureParams]] = SchemaStructureParams

    def inspect(self) -> InspectionResult:
        """
        Inspect synthetic dataset for schema structure, comparing back to
        original source dataset.
        """
        source_data = self.params.source_data
        synth_data = self.params.synth_data

        theme = "Structure"
        name = "check_schema_structure"
        status: str = ""
        message: str = ""

        if set(source_data) != set(synth_data):
            message = "Different tables between source and synthetic data"
            status = "Failed"

            return InspectionResult(
                theme=theme, name=name, status=status, message=message
            )

        else:
            message = "OK"
            status = "Passed"

            return InspectionResult(
                theme=theme, name=name, status=status, message=message
            )



from dataclasses import dataclass, field
from typing import Dict
from pyspark.sql import DataFrame, types as T
from .profiler_column import (
    BaseColumnProfile,
    profile_string_column,
    profile_numerical_column,
    profile_date_column,
)

type_dispatch = {
    T.StringType: profile_string_column,
    T.IntegerType: profile_numerical_column,
    T.LongType: profile_numerical_column,
    T.FloatType: profile_numerical_column,
    T.DoubleType: profile_numerical_column,
    T.DateType: profile_date_column,
}


@dataclass(slots=True)
class TableProfile:
    name: str
    row_count: int
    columns: Dict[str, BaseColumnProfile] = field(default_factory=dict)


def profile_table(df: DataFrame, table_name: str) -> TableProfile:
    col_profiles = {}

    for field in df.schema.fields:
        col_name = field.name
        spark_dtype = field.dataType

        profiler_fn = type_dispatch.get(type(spark_dtype))

        if profiler_fn:
            col_profiles[col_name] = profiler_fn(df, col_name)
        else:
            raise TypeError(
                f"Unsupported column type for '{col_name}': {spark_dtype.typeName()}"
            )

    return TableProfile(name=table_name, row_count=df.count(), columns=col_profiles)



from dataclasses import dataclass, field
from typing import List, Literal

STATUS_OPTIONS = Literal["Passed", "Failed"]


@dataclass
class InspectionResult:
    theme: str
    name: str
    status: STATUS_OPTIONS
    message: str


@dataclass
class InspectionReport:
    results: List[InspectionResult] = field(default_factory=List)



from typing import Dict, Any, List
from dataclasses import dataclass, field, asdict
from pyspark.sql import types as T
from respark.profiling import SchemaProfile


@dataclass
class ColumnGenerationPlan:
    name: str
    data_type: str
    rule: str
    params: Dict[str, Any] = field(default_factory=dict)


@dataclass
class TableGenerationPlan:
    name: str
    row_count: int
    columns: List[ColumnGenerationPlan] = field(default_factory=list)


@dataclass
class SchemaGenerationPlan:
    tables: List[TableGenerationPlan] = field(default_factory=list)

    def to_dict(self) -> Dict[str, Any]:
        return asdict(self)


def make_generation_plan(schema_profile: SchemaProfile) -> SchemaGenerationPlan:
    tables: List[TableGenerationPlan] = []

    for _, table_profile in schema_profile.tables.items():
        col_plans: List[ColumnGenerationPlan] = []
        row_count = table_profile.row_count

        for _, column_profile in table_profile.columns.items():
            col_plans.append(
                ColumnGenerationPlan(
                    name=column_profile.name,
                    data_type=column_profile.normalised_type,
                    rule=column_profile.default_rule(),
                    params=column_profile.type_specific_params(),
                )
            )

        tables.append(
            TableGenerationPlan(
                name=table_profile.name, row_count=row_count, columns=col_plans
            )
        )

    return SchemaGenerationPlan(tables)
